/*
  lircd - Linux Infra-red Remote Control Daemon
  
  Copyright (C) 1996,97 Ralph Metzler (rjkm@thp.uni-koeln.de)
  Copyright (C) 1998 Christoph Bartelmus (columbus@hit.handshake.de)
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published
  by the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 * 0.1:  03/27/96  decode SONY infra-red signals
 *                 create mousesystems mouse signals on pipe /dev/lircm
 *       04/07/96  send ir-codes to clients via socket (see irpty)
 *       05/16/96  now using ir_remotes for decoding
 *                 much easier now to describe new remotes
 *
 * 0.5:  09/02/98 finished (nearly) complete rewrite (Christoph)
 */

#define __USE_BSD

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#ifdef __GLIBC__
#include <sys/perm.h>
#endif
#include <errno.h>

#include "lircd.h"
#include "ir_remote.h"
#include "config.h"

char *progname="lircd";
char *logfile=LOGFILE;

#define HOSTNAME_LEN 128
char hostname[HOSTNAME_LEN+1];

FILE *lf=NULL;
int lirc,sockfd;
int clis[FD_SETSIZE-2]; /* substract one for each lirc and sockfd */
int clin=0;

inline int max(int a,int b)
{
	return(a>b ? a:b);
}

void sigterm(int sig)
{
	int i;
	
	if(last_include_remote!=NULL)
		free_config(last_include_remote->next);
	else
		free_config(remotes);
	logprintf("caugth signal\n");
	for (i=0; i<clin; i++)
	{
		shutdown(clis[i],2);
		close(clis[i]);
	};
	shutdown(sockfd,2);
	close(sockfd);
	close(lirc);
	if(lf) fclose(lf);
	raise(sig);
}

void sighup(int sig)
{
	FILE *fd;
	struct ir_remote *config_remotes;
	
	if(free_remotes!=NULL)
	{
		logprintf("SIGHUP unhandled\n");
		return;
	}
	fd=fopen(CONFIGFILE,"r");
	if(fd==NULL)
	{
		logprintf("could not open config file\n");
		logperror(NULL);
		return;
	}
	config_remotes=read_config(fd);
	fclose(fd);
	if(config_remotes==(void *) -1)
	{
		logprintf("reading of config file failed\n");
	}
	else
	{
		logprintf("config file read\n");
		if(config_remotes==NULL)
		{
			logprintf("warning: config file contains no valid remote control definition\n");
		}
		if(last_include_remote!=NULL)
		{
			/* I cannot free the data structure
			   as they could still be in use */
			free_remotes=last_include_remote->next;
			remotes=include_remotes[0];
			last_include_remote->next=config_remotes;
		}
		else
		{
			free_remotes=remotes;
			remotes=config_remotes;
		}
		if(remotes==NULL)
		{
			logprintf("warning: no valid remote control defined\n");
		}
	}
}

void nolinger(int sock)
{
	static struct linger  linger = {0, 0};
	int lsize  = sizeof(struct linger);
	setsockopt(sock, SOL_SOCKET, SO_LINGER, (void *)&linger, lsize);
}

void remove_client(int fd)
{
	int i;

	for(i=0;i<clin;i++)
	{
		if(clis[i]==fd)
		{
			shutdown(clis[i],2);
			close(clis[i]);
			logprintf("removed client\n");
			
			clin--;
			for(;i<clin;i++)
			{
				clis[i]=clis[i+1];
			}
			return;
		}
	}
#       ifdef DEBUG
	logprintf("internal error in remove_client: no such fd\n");
#       endif
}

void add_client(void)
{
	int fd;
	int clilen;
	struct sockaddr_un client_addr;

	clilen=sizeof(client_addr);
	fd=accept(sockfd,(struct sockaddr *)&client_addr,&clilen);
	if(fd==-1) 
	{
		logprintf("accept() failed\n");
		logperror(NULL);
		exit(EXIT_FAILURE);
	};

	if(fd>=FD_SETSIZE)
	{
		logprintf("connection rejected\n");
		logperror(NULL);
		shutdown(fd,2);
		close(fd);
		return;
	}
	nolinger(fd);
	clis[clin++]=fd;

	logprintf("accepted new client\n");
}

void start_server(void)
{
	struct sockaddr_un serv_addr;
	struct stat s;
	int ret;

	sockfd=socket(AF_UNIX,SOCK_STREAM,0);
	if(sockfd==-1)
	{
		fprintf(stderr,"%s: could not create socket\n",progname);
		perror(progname);
		exit(EXIT_FAILURE);
	};

	/* 
	   get owner, permissions, etc.
	   so new socket can be the same since we
	   have to delete the old socket.  
	*/
	ret=stat(LIRCD,&s);
	if(ret==-1)
	{
		fprintf(stderr,"%s: could not get file information for %s\n",
			progname,LIRCD);
		perror(progname);
		exit(EXIT_FAILURE);
	}
	ret=unlink(LIRCD);
	if(ret==-1)
	{
		fprintf(stderr,"%s: could not delete %s\n",progname,LIRCD);
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	serv_addr.sun_family=AF_UNIX;
	strcpy(serv_addr.sun_path,LIRCD);
	if(bind(sockfd,(struct sockaddr *)&serv_addr,sizeof(serv_addr))==-1)
	{
		fprintf(stderr,"%s; could not assign address to socket\n",
			progname);
		perror(progname);
		exit(EXIT_FAILURE);
	};

	if(chmod(LIRCD,s.st_mode)==-1 || chown(LIRCD,s.st_uid,s.st_gid)==-1)
	{
		fprintf(stderr,"%s: could not set file permissions\n",progname);
		perror(progname);
		exit(EXIT_FAILURE);
	};

	listen(sockfd,3);
	nolinger(sockfd);
  
	lf=fopen(logfile,"a");
	if(lf==NULL)
	{
		fprintf(stderr,"%s: could not open logfile\n",progname);
		perror(progname);
		exit(EXIT_FAILURE);
	}
	gethostname(hostname,HOSTNAME_LEN);
	logprintf("started server socket\n");
}

void logprintf(char *format_str, ...)
{
	time_t current;
	char *currents;
	va_list ap;  

	current=time(&current);
	currents=ctime(&current);
		    
	fprintf(lf,"%15.15s %s %s: ",currents+4,hostname,progname);
	va_start(ap,format_str);
	vfprintf(lf,format_str,ap);
#ifndef DAEMONIZE
	vfprintf(stderr,format_str,ap);
#endif DAEMONIZE
	va_end(ap);
	fflush(lf);
}

void logperror(const char *s)
{
	const char *errmsg;

	if(errno<sys_nerr)
	{
		errmsg=sys_errlist[errno];
	}
	else
	{
		errmsg="unknown error";
	}
		
	if(s!=NULL)
	{
		logprintf("%s: %s\n",s,errmsg);
	}
	else
	{
		logprintf("%s\n",errmsg);
	}
}

#ifdef DAEMONIZE

void daemonize(void)
{
	pid_t pid;
	
	if((pid=fork())<0)
	{
		fprintf(stderr,"%s: fork() failed\n",progname);
		perror(progname);
		exit(EXIT_FAILURE);
	}
	else if(pid) /* parent */
	{
		exit(EXIT_SUCCESS);
	}
	else
	{
		fclose(stdin);
		fclose(stdout);
		fclose(stderr);
		setsid();
		chdir("/");
		umask(0);
	}
}

#endif DAEMONIZE

void sigalrm(int sig)
{
	struct itimerval repeat_timer;

	if(repeat_remote->last_code!=repeat_code)
	{
		/* we received a different code from the original
		   remote control
		   we could repeat the wrong code
		   so better stop repeating */
		return;
	}
	send_command(repeat_remote,repeat_code);

	repeat_timer.it_value.tv_sec=0;
	repeat_timer.it_value.tv_usec=repeat_remote->gap;
	repeat_timer.it_interval.tv_sec=0;
	repeat_timer.it_interval.tv_usec=0;

	setitimer(ITIMER_REAL,&repeat_timer,NULL);
}

void get_command(int fd)
{
	int length;
	char buffer[PACKET_SIZE+1];
	char *end;
	int packet_length;
	char *directive,*name,*command;
	struct ir_remote *remote;
	struct ir_ncode *code;
	struct itimerval repeat_timer,test;

	test=repeat_timer;
	length=read(fd,buffer,PACKET_SIZE);
	packet_length=0;
	while(length>packet_length)
	{
		buffer[length]=0;
		end=strchr(buffer,'\n');
		if(end==NULL)
		{
			logprintf("bad send packet: \"%s\"\n",buffer);
			while(length>0)
			{
				length=read(fd,buffer,PACKET_SIZE);
			}
			return;
		}
		end[0]=0;
#               ifdef DEBUG
		logprintf("sending: \"%s\"\n",buffer);
#               endif		
		packet_length=strlen(buffer)+1;

		directive=strtok(buffer,WHITE_SPACE);
		name=strtok(NULL,WHITE_SPACE);
		command=strtok(NULL,WHITE_SPACE);
		
		if(directive==NULL || name==NULL || command==NULL || 
		   strtok(NULL,WHITE_SPACE)!=NULL)
		{
			logprintf("bad send packet\n");
			return;
		}
		remote=get_ir_remote(name);
		if(remote==NULL)
		{
			logprintf("unknown remote: \"%s\"\n",name);
			return;
		}
		code=get_ir_code(remote,command);
		if(code==NULL)
		{
			logprintf("unknown command: \"%s\"\n",command);
			return;
		}
		if(strcasecmp(directive,"SEND_ONCE")==0)
		{
			if(remote==repeat_remote)
			{
				logprintf("remote is repeating\n");
				return;
			}
			if(remote->repeat_bit>0)
				remote->repeat_state=
				!remote->repeat_state;

			send_command(remote,code);
		}
		else if(strcasecmp(directive,"SEND_START")==0)
		{

			if(repeat_remote!=NULL)
			{
				logprintf("already repeating\n");
			}
			else
			{
				repeat_remote=remote;
				repeat_code=code;

				if(remote->repeat_bit>0)
					remote->repeat_state=
					!remote->repeat_state;
				send_command(remote,code);

				repeat_timer.it_value.tv_sec=0;
				repeat_timer.it_value.tv_usec=remote->gap;
				repeat_timer.it_interval.tv_sec=0;
				repeat_timer.it_interval.tv_usec=0;
				setitimer(ITIMER_REAL,&repeat_timer,NULL);
			}
		}
		else if(strcasecmp(directive,"SEND_STOP")==0)
		{
			if(repeat_remote && repeat_code &&
			   strcasecmp(name,repeat_remote->name)==0 && 
			   strcasecmp(command,repeat_code->name)==0)
			{
				repeat_timer.it_value.tv_sec=0;
				repeat_timer.it_value.tv_usec=0;
				repeat_timer.it_interval.tv_sec=0;
				repeat_timer.it_interval.tv_usec=0;
				
				setitimer(ITIMER_REAL,&repeat_timer,NULL);
				
				repeat_remote=NULL;
				repeat_code=NULL;
			}
			else
			{
				logprintf("not repeating\n");
			}
		}
		else
		{
			logprintf("unknown directive \"%s\"\n",directive);
		}
		if(length>packet_length)
		{
			int new_length;

			memmove(buffer,buffer+packet_length,
				length-packet_length+1);
			if(strchr(buffer,'\n')==NULL)
			{
				new_length=read(fd,buffer+length-packet_length,
						PACKET_SIZE-(length-packet_length));
				if(new_length>0)
				{
					length=length-packet_length+new_length;
				}
				else
				{
					length=new_length;
				}
			}
			else
			{
				length-=packet_length;
			}
		}
	}

	if(length==0) /* EOF: connection closed by client */
	{
		remove_client(fd);
	}
	else if(length==-1)
	{
		logprintf("read failed\n");
	}
}

/* A safer write(), since sockets might not write all but only some of the
   bytes requested */

inline int write_socket(int fd, char *buf, int len)
{
	int done,todo=len;

	while(todo)
	{
		done=write(fd,buf,todo);
		if (done<=0) return(done);    
		buf+=done;
		todo-=done;
	}
	return(len-todo);
}

void free_old_remotes()
{
	struct ir_remote *scan_remotes,*found;
	struct ir_ncode *code;

	if(last_remote!=NULL)
	{
		scan_remotes=free_remotes;
		while(scan_remotes!=NULL)
		{
			if(last_remote==scan_remotes)
			{
				found=get_ir_remote(last_remote->name);
				if(found!=NULL)
				{
					code=get_ir_code(found,last_remote->last_code->name);
					if(code!=NULL)
					{
						found->reps=last_remote->reps;
						found->repeat_state=last_remote->repeat_state;
						last_remote=found;
						last_remote->last_code=code;
					}
				}
				break;
			}
			scan_remotes=scan_remotes->next;
		}
		if(scan_remotes==NULL) last_remote=NULL;
	}
	/* check if last config is still needed */
	found=NULL;
	if(repeat_remote!=NULL)
	{
		scan_remotes=free_remotes;
		while(scan_remotes!=NULL)
		{
			if(repeat_remote==scan_remotes)
			{
				found=repeat_remote;
				break;
			}
			scan_remotes=scan_remotes->next;
		}
		if(found!=NULL)
		{
			found=get_ir_remote(repeat_remote->name);
			if(found!=NULL)
			{
				code=get_ir_code(found,repeat_code->name);
				if(code!=NULL)
				{
					struct itimerval repeat_timer;

					repeat_timer.it_value.tv_sec=0;
					repeat_timer.it_value.tv_usec=0;
					repeat_timer.it_interval.tv_sec=0;
					repeat_timer.it_interval.tv_usec=0;

					found->last_code=code;
					found->last_send=repeat_remote->last_send;
					found->repeat_state=repeat_remote->repeat_state;
					
					setitimer(ITIMER_REAL,&repeat_timer,&repeat_timer);
					/* "atomic" */
					repeat_remote=found;
					repeat_code=code;
					/* end "atomic" */
					setitimer(ITIMER_REAL,&repeat_timer,NULL);
					found=NULL;
				}
			}
			else
			{
				found=repeat_remote;
			}
		}
	}
	if(found==NULL && decoding!=free_remotes)
	{
		free_config(free_remotes);
		free_remotes=NULL;
	}
#       ifdef DEBUG
	else
	{
		logprintf("free_remotes still in use\n");
	}
#       endif
}


unsigned long readdata()
{
	fd_set fds;
	int maxfd,i,ret;
  
	while(1)
	{
		FD_ZERO(&fds);
		FD_SET(lirc,&fds);
		FD_SET(sockfd,&fds);
		maxfd=max(lirc,sockfd);
#if defined(DEBUG) && !defined(DAEMONIZE)
		FD_SET(STDIN_FILENO,&fds);
		maxfd=max(maxfd,STDIN_FILENO);
#endif

		for(i=0;i<clin;i++)
		{
			FD_SET(clis[i],&fds);
			maxfd=max(maxfd,clis[i]);
		}
		
		/*
		  select() is still interrupted despite of SA_RESTART
		  Is this a bug or desired behaviour ???
		*/
		
		do{
			do{
				ret=select(maxfd+1,&fds,NULL,NULL,NULL);
				if(free_remotes!=NULL)
				{
					free_old_remotes();
				}
			}
			while(ret==-1 && errno==EINTR);
			if(ret==-1)
			{
				logprintf("select failed\n");
				logperror(NULL);
				continue;
			}
		}
		while(ret==-1);
		
		for(i=0;i<clin;i++)
		{
			if(FD_ISSET(clis[i],&fds))
			{
				FD_CLR(clis[i],&fds);
				get_command(clis[i]);
			}
		}
		if(FD_ISSET(sockfd,&fds))
		{
#                       ifdef DEBUG
			logprintf("registering new client\n");
#                       endif
			add_client();
		}
		if(FD_ISSET(lirc,&fds))
		{
			unsigned long data;
			int ret;
			
			do
			{
				ret=read(lirc,&data,sizeof(unsigned long));
#                               ifdef DEBUG
				if(ret!=sizeof(unsigned long))
				{
					logprintf("error reading from lirc\n");
					logperror(NULL);
				}
#                               endif
			}
			while(ret!=sizeof(unsigned long));
			
			return(data);
		}
#if defined(DEBUG) && !defined(DAEMONIZE)
		if(FD_ISSET(STDIN_FILENO,&fds))
		{
			unsigned long data;
			int ret;

			ret=fscanf(stdin,"space %ld\n",&data);
			if(ret==1)
			{
				return(data);
			}
			ret=fscanf(stdin,"pulse %ld\n",&data);
			if(ret==1)
			{
				return(PULSE_BIT|data);
			}
			ret=fscanf(stdin,"%*s\n");
			if(ret==EOF)
				exit(0);
		}
#endif
	}
}

void loop()
{
	char *message;
	int len,i;
	unsigned long data;

	while(1)
	{
		data=readdata();
		
		message=decode_command(data);

		if(message!=NULL)
		{
			len=strlen(message);
				
			for (i=0; i<clin; i++)
			{
#                               ifdef DEBUG
				logprintf("writing to client %d\n",i);
#                               endif
				if(write_socket(clis[i],message,len)<len)
				{
					remove_client(clis[i]);
					i--;
				}				
			}
		}
	}
}  

int main(int argc,char **argv)
{
	struct sigaction act;
	int i;

	/* link included remotes */
	last_include_remote=NULL;
	for(i=0;include_remotes[i]!=NULL;)
	{
		last_include_remote=include_remotes[i];
		include_remotes[i]->next=include_remotes[++i];
	}
	remotes=include_remotes[0];
#       if 0
	fprint_remotes(stdout,remotes);
	fflush(stdout);
#       endif

	if((lirc=open("/dev/lirc",O_RDWR))<0)
	{
		fprintf(stderr,"%s: could not open lirc\n",progname);
		perror(progname);
		exit(EXIT_FAILURE);
	}

	signal(SIGPIPE,SIG_IGN);
	signal(SIGTERM,sigterm);
	signal(SIGINT,sigterm);

	act.sa_handler=sigalrm;
	sigemptyset(&act.sa_mask);
	act.sa_flags=SA_RESTART;           /* don't fiddle with EINTR */
	sigaction(SIGALRM,&act,NULL);

	start_server();

	act.sa_handler=sighup;
	sigemptyset(&act.sa_mask);
	act.sa_flags=SA_RESTART;           /* don't fiddle with EINTR */
	sigaction(SIGHUP,&act,NULL);

	raise(SIGHUP);                     /* read config file */

#ifdef DAEMONIZE
	/* ready to accept connections */
	daemonize();
#endif

#if defined(DEBUG) && !defined(DAEMONIZE)
	if(argc>1) sleep(argc);
#endif

	loop();

	/* never reached */
	return(EXIT_SUCCESS); 
}
