/*
   lirc -  Linux Infrared Remote Control
   device driver that records pulse- and pause-lengths
   between DDCD event
   it also writes 

   Copyright (C) 1996,97 Ralph Metzler (rjkm@thp.uni-koeln.de)
   Copyright (C) 1998 Trent Piepho <xyzzy@u.washington.edu>

   lirc is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published
   by the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   lirc is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define __KERNEL__
#define MODULE
#include <linux/version.h>
#include <linux/module.h>

#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/config.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/serial_reg.h>
#include <linux/signal.h>
#include <linux/time.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/wait.h>
#include <linux/mm.h>

#include <linux/delay.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/fcntl.h>

/* A long pulse code from a remote might take upto 300 bytes.  The daemon
 * should read the bytes as soon as they are generated, so take the number of
 * keys you think you can push before the daemon runs and multiply by 300. 
 * The driver will warn you if you overrun this buffer.  If you have a slow
 * computer or non-busmastering IDE disks, maybe you will need to increase
 * this.  */
/* This MUST be a power of two!  It has to be larger than 16 as well. */
#define RBUF_LEN 1024
#define WBUF_LEN 1024


#define IR_SOFTCARRIER 1


static int major = LIRC_MAJOR;
static int mode = 2;
static int sense = -1;  /* -1 = auto, 0 = active high, 1 = active low */
static int flags = IR_SOFTCARRIER;


static struct wait_queue *lirc_wait_in = NULL;

static int port = 0x2f8;
static int irq = 3;

static unsigned char sendon = UART_MCR_DTR |UART_MCR_RTS | UART_MCR_OUT2;
static unsigned char sendoff= UART_MCR_RTS | UART_MCR_OUT2;

static unchar rbuf[RBUF_LEN];
static int rbh, rbt;

static unsigned long wbuf[WBUF_LEN];

static inline unsigned int sinp(int offset)
{
    return inb(port + offset);
}

static inline void soutp(int offset, int value)
{
    outb(value, port + offset);
}

void on(void) {
    soutp(UART_MCR, sendon);
}
  
void off(void) {
    soutp(UART_MCR, sendoff);
}

void send_pulse(unsigned long length) {
    unsigned int k;
    if (!length)
	return;
    if (flags&IR_SOFTCARRIER) {
	length=(length+12)/25;
	for (k=0; k<length; k++) {
	    on();
	    udelay(13);
	    off();
	    udelay(12);
	}
    } else {
	on();
	udelay(length);
    }
}

void send_space(ulong length) {
    if (!length)
	return;
    off();
    udelay(length);
}

static inline void rbwrite(ulong l)
{
    *(ulong*)(rbuf+rbt) = l;
    rbt = (rbt + 4) & (RBUF_LEN-1);
    if (rbt == rbh)  {
		printk(KERN_WARNING "lirc: Buffer overrun\n");
        rbh = (rbh + 4) & (RBUF_LEN-1);
	};
}

void irq_handler(int i, void *blah, struct pt_regs *regs)
{
    struct timeval tv;
    int status;
    ulong deltv;
    static struct timeval lasttv={0,0};
    static int lastdcd=0;

    status = sinp(UART_MSR);
    if(MOD_IN_USE && (status&UART_MSR_DDCD))  {
        /* get current time */
        do_gettimeofday(&tv);

        if(mode==2)  {
            /* new mode, written by Trent Piepho <xyzzy@u.washington.edu> */
            /* This has a 4 byte packet format.  The MSB is a one byte value
             * that is either 0, meaning space, or 1, meaning pulse.  The
             * driver needs to know if your receiver is active high or active
             * low, or the space/pulse sense could be inverted.  The lower
             * three bytes are the length in microseconds.  Lengths greater
             * than or equal to 16 seconds are clamped to 0xffffff.  This is a
			 * much simpler interface for user programs, as well as
			 * eliminating "out of phase" errors with space/pulse
			 * autodetection.  */

            /* calculate time since last interrupt in microseconds */
            deltv = tv.tv_sec - lasttv.tv_sec;
            if(deltv > 15)  {
                deltv = 0xFFFFFF;  /* really long time */
            } else {
                deltv = deltv*1000000 + tv.tv_usec - lasttv.tv_usec;
            };
            rbwrite(lastdcd?(deltv|0x1000000):deltv);
            lasttv = tv;
            lastdcd = status&UART_MSR_DCD;
            if(sense) lastdcd=!lastdcd;
        } else if(mode==1)  {
            /* Just write the timeval of each interrupt to buffer */
			printk(KERN_ERR "lirc: Mode 1 isn't supported anymore!\n");
        } else {
            /* calculate time since last interrupt in microseconds */
			printk(KERN_ERR "lirc: Mode 0 isn't supported anymore!\n");
        }
        wake_up_interruptible(&lirc_wait_in);
    }
    sinp(UART_IIR);
}

static int init_port()
{
    int result;

    result = request_irq(irq, irq_handler, SA_INTERRUPT, "lirc", NULL);
    switch (result) {
    case -EBUSY:
        printk(KERN_ERR "lirc: IRQ %d busy\n", irq);
        return -EBUSY;
    case -EINVAL:
        printk(KERN_ERR "lirc: Bad irq number or handler\n");
        return -EINVAL;
    default:
        break;
    };

    printk(KERN_INFO "lirc: Interrupt %d, port %04x obtained\n", irq, port);

    /* reserve io region */
    request_region(port,8,"lirc");

    cli();

    /* init read buffer pointers */
    rbh = rbt = 0;

    /* set DLAB 0 */
    soutp(UART_LCR, sinp(UART_LCR) & 0x7f);

    /*clear registers */
    sinp(UART_LSR);
    sinp(UART_RX);
    sinp(UART_IIR);
    sinp(UART_MSR);

    /* set line for power source and enable interrupts */
    /*soutp(UART_MCR, UART_MCR_DTR | UART_MCR_RTS | UART_MCR_OUT2);*/
    soutp(UART_MCR, UART_MCR_RTS | UART_MCR_OUT2);
    soutp(UART_IER, UART_IER_MSI);

    /*clear registers again to be sure */
    sinp(UART_LSR);
    sinp(UART_RX);
    sinp(UART_IIR);
    result = sinp(UART_MSR);
    sti();
    /* If DCD is high, then this must be an active low receiver */
	if(sense==-1)  {
		sense=(result&UART_MSR_DCD);
		printk(KERN_INFO "lirc: auto-detected active %s receiver\n",
			sense?"low":"high");
	} else {
		printk(KERN_INFO "lirc: Manually using active %s recevier\n",
			sense?"low":"high");
	};
    return 0;
}

static int lirc_open(struct inode *ino, struct file *filep)
{
    unsigned int minor = MINOR(ino->i_rdev);

    if (MOD_IN_USE) return -EBUSY;
    printk(KERN_DEBUG "lirc: open minor = %d\n", minor);
    MOD_INC_USE_COUNT;
    return 0;
}

static void lirc_close(struct inode *node, struct file *file)
{
    MOD_DEC_USE_COUNT;
}

static int lirc_select(struct inode *node, struct file *file,
                       int sel_type, select_table * wait)
{
    if (sel_type != SEL_IN)
        return 0;
    if (rbh != rbt)
        return 1;
    select_wait(&lirc_wait_in, wait);
    return 0;
}

#ifdef KERNEL_2_1
static long lirc_read(struct inode *node, struct file *file, char *buf,
                      unsigned long count)
#else
static int lirc_read(struct inode *node, struct file *file, char *buf,
                     int count)
#endif
{
    int n = 0, retval = 0;

    while (n < count) {
        cli();
        if (rbt != rbh) {
            memcpy_tofs((void *) buf + n, (void *) (rbuf + rbh), 1);
            rbh = (rbh + 1) & (RBUF_LEN - 1);
            n++;
            sti();
        } else {
            sti();
            if (file->f_flags & O_NONBLOCK) {
                retval = -EAGAIN;
                break;
            }
            if (current->signal & ~current->blocked) {
                retval = -EINTR;
                break;
            }
            interruptible_sleep_on(&lirc_wait_in);
            current->state = TASK_RUNNING;
        }
    }
    return (n ? n : retval);
}

#ifdef KERNEL_2_1
static long lirc_write(struct inode *node, struct file *file, const char *buf,
                      unsigned long n)
#else
static int lirc_write(struct inode *node, struct file *file, const char *buf,
                     int n)
#endif
{
	int retval,i,count;
	
	if(n%sizeof(unsigned long)) return(-EINVAL);
	
	retval=verify_area(VERIFY_READ,buf,n);
	if(retval) return(retval);

	count=n/sizeof(unsigned long);

	if(count>WBUF_LEN || count%2==0) return(-EINVAL);

	memcpy_fromfs(wbuf,buf,n);

	for(i=0;i<count;i++)
	{
		if(i%2) send_space(wbuf[i]);
		else send_pulse(wbuf[i]);
	}
	off();
	return(n);
}

static struct file_operations lirc_fops =
{
    NULL,                       /* lseek */
    lirc_read,                  /* read */
    lirc_write,                 /* write */
    NULL,                       /* readdir */
    lirc_select,                /* select */
    NULL,						/* ioctl */
    NULL,                       /* mmap  */
    lirc_open,					/* open */
    lirc_close,					/* release */
    NULL,                       /* fsync */
    NULL,                       /* fasync */
    NULL,                       /* check_media_change */
    NULL                        /* revalidate */
};

int init_module(void)
{
	int result;

    if ((result = init_port()) < 0)
        return result;
    if (register_chrdev(major, "lirc", &lirc_fops) < 0) {
        printk(KERN_ERR "lirc: register_chrdev failed!\n");
        return -EIO;
    }
    printk(KERN_INFO "lirc: Successfully installed!\n");
    return 0;
}

void cleanup_module(void)
{
    if (MOD_IN_USE) return;
    free_irq(irq, NULL);
    printk(KERN_INFO "lirc: freed IRQ %d\n", irq);
    release_region(port,8);
    unregister_chrdev(major, "lirc");
    printk(KERN_INFO "lirc: cleaned up module\n");
}
